/*
SAS Scientific Discovery Solutions 

Name: InputEngineMacros
Author: Tzu-Ming Chu, SAS Institute Inc.
Support: Tzu-Ming Chu, SAS Institute Inc.
History:
   tzuchu 28Aug02 initial coding

Description: InputEngineMacros contains
a collection of SAS macros that can be 
called from Input Engine.  
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

%macro DesignParse;
*** load in experiment design;
%if %upcase(&ExpType) = SAS %then %do;
   %PathName(&ExpInFile);
   libname tmp "&TmpPath";
   libname library "&TmpPath";
   proc datasets lib=OutLib nolist;
      %if %sysfunc(exist(OutLib.exptable)) %then delete exptable;;
      copy in=tmp out=OutLib;
         select &TmpName;
      run;
	  %if &TmpName ne exptable %then change &TmpName=exptable; ;
   quit;
%end;
%else %do;
   * Check Excel file and OS system;
   %CheckExcel(&ExpType,The Experimental Design Table);
   %if &exiterror %then %goto exitdesignparse;
   ***** Added optional delimiter statement to handle bogus end-of-line characters;
   %if %upcase(&ExpType) = CSV %then %let delimiter = DELIMITER='2C0D'x;
   %else %if %upcase(&ExpType) = DLM %then %let delimiter = DELIMITER='200D'x;
   %else %if %upcase(&ExpType) = TAB %then %let delimiter = DELIMITER='090D'x;
   %else %let delimiter = ;
   PROC IMPORT OUT= OutLib.exptable 
        DATAFILE= "&ExpInFile"
        DBMS=&ExpType REPLACE;
      GETNAMES=YES;
      &delimiter;
   RUN; 
%end;

* screen out empty column to avoid empty or missing column(s) in final merged data;
%let tid = %sysfunc(open(OutLib.exptable));
%let nObs = %sysfunc(attrn(&tid,NOBS));
%let nVars = %sysfunc(attrn(&tid,NVARS)); 
%do i = 1 %to &nVars;
   %let CK_Var&i = %sysfunc(varname(&tid,&i));
   %let CK_VarType&i = %sysfunc(vartype(&tid,&i));
%end;
%let tid = %sysfunc(close(&tid));

data _null_;
   set OutLib.exptable end=eof;
   array flag{&nVars} (&nVars*0);
   retain flag1-flag&nVars;  
   %do i = 1 %to &nVars;
      %if &&CK_VarType&i=C %then if &&CK_Var&i ne '' then flag{&i} = 1;
	  %else if &&CK_Var&i ne . then flag{&i} = 1; ;
   %end;
   if eof;
   SumFlag = Sum( of flag1-flag&nVars ); 
   Call symput('SumFlag',SumFlag);
   %do i = 1 %to &nVars;
      call symput("Flag&i",flag&i);
   %end; 
run;

%if %eval(&SumFlag-&nVars) %then %do;
   %let DelVars = ;
   %do i = 1 %to &nVars;
       %if &&Flag&i = 0 %then %let DelVars = &DelVars &&CK_Var&i;
   %end;
   %if %upcase(%substr(&DataPlatForm,1,4))=AFFY %then %do;
      data OutLib.exptable;
         set OutLib.exptable;
	     drop &DelVars;
      run;
   %end;
   %else %do;
      proc sort data=OutLib.exptable(drop=&DelVars);
	     by File;
	  run;
   %end;
%end;

%global Fnum;

* Open the new design table and get some stats;
%let tid = %sysfunc(open(OutLib.exptable));
%let Fnum = %sysfunc(varnum(&tid,FILE));
%let nVars = %sysfunc(attrn(&tid,NVARS));

   * Adjust for SAS naming conventions;
   %global XnVars AllXVars;
   %let XnVars = %eval(&nVars-&Fnum);
   %let AllXVars = ;
   %do i = 1 %to &XnVars; 
      %global XVar&i;
	  %let XVar&i = %sysfunc(varname(&tid,%eval(&i+&Fnum)));
	  %if %upcase(&&XVar&i) ne INTENSITY %then %let AllXVars = &AllXVars &&XVar&i;
   %end;
   * Close the design table;
   %let tid = %sysfunc(close(&tid));
   %if %upcase(%substr(&DataPlatForm,1,4)) ne AFFY %then %let XVars = Intensity &AllXVars;
   %else %let XVars = &AllXVars;
   %Naming(Dset=outlib.exptable,nVars=&XnVars,AllVars=&Xvars);
 
*check variable names in experimental design table;
%let tid = %sysfunc(open(OutLib.exptable));
%global AllVars;
%let AllVars = ;
%do i = 1 %to %eval(&Fnum-1);
   %global VarName&i VarType&i;
   %let VarName&i = %sysfunc(varname(&tid,&i));
   %let VarType&i = %sysfunc(vartype(&tid,&i));
   %let AllVars = &AllVars &&VarName&i;
%end;

* Close the design table;
%let tid = %sysfunc(close(&tid));
%put DesignParse completed.;

%exitdesignparse:;
%if (&syserr | &exiterror) %then %put ERROR: DesignParse exited due to errors.;
%mend DesignParse;


****************************************;
*****   DatParse imports           *****;
*****      1. Agilent .TXT         *****;
*****      2. GenPix .GPR          *****;
*****      3. QuantArray .TXT      *****;
*****      4. ScanAnlyze .DAT      *****;
****************************************;
%macro DATParse;
%let VeryStart = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%let Day1 = %sysfunc(round(%sysevalf(%sysfunc(date()))));

data ETable;
   set OutLib.exptable;
   if file ne '';
run;

%* start parsing;
%if %sysfunc(exist(OutLib.&OutData)) %then 
   %do;
      proc datasets library=OutLib nolist; 
         delete &OutData spotmap; 
      quit;
   %end;

%global nRawFiles RawFiles;
data ETable;
   set ETable;
   n = _n_;
run;
proc sort data=ETable(keep=File n) out=tmp nodupkey;
   by File;
run;
proc sort data=tmp(drop=file);
   by n;
run;
data ETable;
   merge ETable tmp(in=in);
   by n;
   if in then Flag = 1;
   drop n;
run;
%nobsvars(tmp);
%let nRawFiles = &nObs;
%let nRec = 0;

%nobsvars(ETable);
%let nRows = &nObs;
%let nCols = %eval(&nVars-1);
%if %upcase(&Log2Transform) = YES %then %let KeepI = Log2i;
                                  %else %let KeepI = Intensity;
%let NameExist = 0; * indicator of the existence of gene name in QuantArray file;
%do i_row = 1 %to &nRows; 
   %let FileFlag = 0;
   data _null_;
      set ETable;
      if _n_ = &i_row;
	  call symput('Dat',trim(left(File)));
	  call symput('Int',trim(left(Intensity)));
      %do j = 1 %to %eval(&nCols-&XnVars-1);
	     call symput("V&j",&&VarName&j);
	     call symput("T&j","&&VarType&j");
      %end;
	  %do k = 1 %to &XnVars;
	     call symput("AddVar&k",&&XVar&k);
	  %end;
	  if Flag then call symput("FileFlag",1);
   run;
   
   /* check if file exists */
   %let tmp_flag = 0;
   %if (%upcase(&Dataplatform)=AGILENT & %upcase(%substr(%trim(&Dat),%length(%trim(&Dat))-3))=.TXT) |
       (%upcase(&Dataplatform)=GENEPIX & %upcase(%substr(%trim(&Dat),%length(%trim(&Dat))-3))=.GPR) |
       (%upcase(&Dataplatform)=QUANTARRAY & %upcase(%substr(%trim(&Dat),%length(%trim(&Dat))-3))=.TXT) |
       (%upcase(&Dataplatform)=SCANALYZE & %upcase(%substr(%trim(&Dat),%length(%trim(&Dat))-3))=.DAT) 
      %then %let tmp_flag = 1;
   %if &tmp_flag %then %do;
      %let DataFile = &DataInPath.&Dat;
	  %if %sysfunc(fileexist(&DataFile)) = 0 %then %put ERROR: &DataInPath.&DAT does not exist.;
   %end;
   %else %do;
      %if %upcase(&Dataplatform)=AGILENT %then %let extension = txt;
	  %else %if %upcase(&Dataplatform)=GENEPIX %then %let extension = gpr;
	  %else %if %upcase(&Dataplatform)=QUANTARRAY %then %let extension = txt;
	  %else %if %upcase(&Dataplatform)=SCANALYZE %then %let extension = dat;
      filename DataFil1 "&DataInPath.&DAT..%upcase(%trim(&extension))";
      filename DataFil2 "&DataInPath.&DAT..%trim(&extension)";
	  %if %sysfunc(fexist(DataFil1)) %then %do;
         %let datafile=&DataInPath.&DAT..%upcase(%trim(&extension));
		 %let DAT = &DAT..%upcase(%trim(&extension));
	  %end;
      %else %if %sysfunc(fexist(DataFil2)) %then %do;
         %let datafile=&DataInPath.&DAT..%trim(&extension);
		 %let DAT = &DAT..%trim(&extension);
	  %end;
	  %else %put ERROR: &DataInPath.&DAT does not exist.;
   %end;

%if &FileFlag %then %do; * avoid duplicate importing;
   /* import DAT files */ 
   %if %upcase(&Dataplatform)=AGILENT %then %do;
	  DATA _null_;
  	     INFILE "&Datafile" MISSOVER DSD  lrecl=32767 firstobs=1 OBS=100;
  		 LENGTH var1 $ 100;
  		 input var1;
  		 if substr(var1,1,7)='FEATURE' then call symput('VRow',_n_);
      run;
   	  %skip_header_import(Datafile=&Datafile,firstrow=%eval(&vrow+1),VarnameRow=&vrow); 
   %end;
   %else %if %upcase(&Dataplatform)=GENEPIX %then %do;
      DATA _null_;
  	  	 INFILE "&Datafile" MISSOVER DSD  lrecl=32767 firstobs=1 OBS=100;
  		 LENGTH var1 $ 100;
  		 input var1;
  		 if substr(var1,1,7)='"Block"' then call symput('VRow',_n_);
      run;
      %skip_header_import(Datafile=&Datafile,firstrow=%eval(&vrow+1),VarnameRow=&vrow); 
   %end;
   %else %if %upcase(&Dataplatform)=QUANTARRAY %then %do;
      %let Encoding = ;
      %if %upcase(&DataEncoding) = UNICODE %then %do;
         %if &sysscp=WIN %then %let Encoding = Encoding=ucs2;
         %else %let Encoding = Encoding=U_L_UNICODE_CE;
      %end;
      /* importing QuantArray raw data */
      data t; * locate first and last observations; 
         infile "&Datafile" &encoding Missover Truncover lrecl=30;
         length x $20;
         input x 1-20;
         if upcase(substr(x,1,10)) = 'BEGIN DATA' then call symput('FirstLine',_n_+2);
         if upcase(substr(x,1,8)) = 'END DATA' then do;
            call symput('LastLine', _n_-1); 
	        stop;
         end;
      run;
      
      data _null_; * get variable names;
         infile "&Datafile" &encoding Missover Truncover
                lrecl=2000 firstobs=%eval(&firstline-1);
         length x $2000 y $2000;
		 input x 1-2000;
         y = translate(compbl(translate(left(trim(x)),'','.')),'_','');
         i = index(y,'Name');
         if i then do;
            z = substr(y,1,41)||'Name $'||substr(y,46);
			call symput('NameExist',1);
			call symput('NameStart',i);
		 end;
         call symput('varnames',z);
		 stop;
      run;
      %let TotalObs = %eval(&LastLine-&FirstLine+1);
      %let LastVar=%scan(&varnames,-1);  
	  %let ckRows = 200;
	  %let DelName = 0;
      data _null_; * chack if gene name is empty;
         infile "&Datafile" &encoding Missover Truncover DLM='090d'X lrecl=2000 firstobs=&firstline;
         retain Flag 0;
         if _n_<=&ckRows then do; 
            input &varnames;
	        if &LastVar ne . then do;
	           Flag = 1;
		       stop;
	        end;
	        if _n_ = &ckRows & ^Flag then do;
	           call symput('DelName',&NameStart);
               stop;
	        end;
         end;
      run;

      %if &DelName %then %let VarNames = %substr(&VarNames,1,%eval(&DelName-1)) %substr(&VarNames,%eval(&DelName+7));
      data Dat;
         infile "&Datafile" &encoding Missover Truncover DLM='090d'X lrecl=2000 firstobs=&firstline;
		 %if ^&DelName %then length Name $100; ; * assign length 100 to gene name;
         if _n_<=&totalobs then do; 
            input &varnames;
            output;
         end; 
      run;
   %end;
   %else %if %upcase(&Dataplatform)=SCANALYZE %then %do;
      %if &i_row = 1 %then %do;
         %let VarList = ;
	     %let LenList = ;
         %skip_header_import_2(DataFile=%quote(&DataFile),DataType=TAB,OutSet=Dat,FirstRow=&DataStartRow,
            VarNameRow=1,GuessRows=&ScanRows,_varlist=VarList,_lenlist=LenList);
	  %end;
	  %else %do;
         data DAT;
            infile "&datafile" DELIMITER='090d'x MISSOVER DSD lrecl=32767 firstobs=&DataStartRow;
            length &LenList;
            input &VarList;
         run;
	  %end;
   %end;
   %if %upcase(&UploadRawFiles) = YES %then %do;
      %let nRec = %eval(&nRec+1);
      %if &nRec = 1 %then %let RawFiles = &DAT;
      %else %let RawFiles = &RawFiles,&DAT;
   %end;
%end;   

   %if &i_row = 1 %then %do;
      %if %upcase(&Dataplatform)=AGILENT %then %do;
		 %let OldVList = FeatureNum Row Col ProbeUID ControlType ProbeName GeneName 
                         SystematicName Description PositionX PositionY;
		 %let CoorVList = ;
		 %checkcoorvar(_ds=Dat,_Old=&OldVList,_New=CoorVList);
         data OutLib.spotmap;
	        set Dat(keep=&CoorVList);
			if FeatureNum=. and PositionX=. and PositionY=. then delete;
	     run;
      %end;
      %if %upcase(&Dataplatform)=GENEPIX %then %do;
		 %let OldVList = Block Column Row Name ID X Y;
		 %let CoorVList = ;
		 %checkcoorvar(_ds=Dat,_Old=&OldVList,_New=CoorVList);
         data OutLib.spotmap;
	        set Dat(keep=&CoorVList);
			if id="" and x=. and y=. then delete;
	     run;
      %end;
      %if %upcase(&Dataplatform)=QUANTARRAY %then %do;
	     %if ^&DelName %then %let tmp = Name;
		 %else %let tmp = ;  
		 %let OldVList = Number Array_Row Array_Column Row Column &tmp 
                          X_Location Y_Location;
		 %let CoorVList = ;
		 %checkcoorvar(_ds=Dat,_Old=&OldVList,_New=CoorVList);
         data OutLib.spotmap;
	        set Dat(keep=&CoorVList);
	     run;
      %end;
      %if %upcase(&Dataplatform)=SCANALYZE %then %do;
		 %let OldVList = SPOT GRID TOP LEFT BOT RIGHT ROW COL;
		 %let CoorVList = ;
		 %checkcoorvar(_ds=Dat,_Old=&OldVList,_New=CoorVList);
         data OutLib.spotmap;
	        set Dat(keep=&CoorVList);
	     run;
      %end;

	  * rename the variables in spotmap if there are corresponding names in design table;
      %if %eval(&XnVars-1) %then %do;
	     %let RenameList = ;
         %let tid = %sysfunc(open(ETable));
         %let rc=%sysfunc(fetchobs(&tid,1));
         %do i = 1 %to %eval(&XnVars-1);
            %let vn=%sysfunc(varnum(&tid,%scan(&AllXVars,&i)));
            %let varc=%sysfunc(getvarc(&tid,&vn));
	        %let j = 1;
	        %do %while(%length(%scan(&CoorVList,&j)));
	           %if %index(%upcase(%scan(&CoorVList,&j)),%upcase(&varc)) &
                   %upcase(&varc)^=%upcase(%scan(&AllXVars,&i)) %then
		          %let RenameList = &RenameList %scan(&CoorVList,&j)=%scan(&AllXVars,&i);
		       %let j = %eval(&j+1);
	        %end;
         %end;
         %let tid = %sysfunc(close(&tid));
	     %if %length(&RenameList) %then %do;
	        proc datasets library=OutLib nolist;
		       modify spotmap;
		       rename &RenameList;
            quit;
	     %end;
      %end;
   %end;

   data TmpDAT;
      set DAT(rename=(&int=_tmpint));
	  &int= _tmpint+0; 
	  %* copy the values of experimental factors;
	  %do j = 1 %to %eval(&nCols-&XnVars-1);
	     %if &&T&j = N %then %do; &&VarName&j = &&V&j; %end;
	                   %else %do; &&VarName&j = "&&V&j"; %end;
	  %end;
	  %if %upcase(&Log2Transform) = YES %then %do;
	     if &Int>0 then log2i = log2(&Int);
	               else log2i = .;
	  %end;
	  %else %do;
	     %if %trim(%left(%upcase(&int))) ne INTENSITY %then Intensity = &Int; ;
	  %end;
	  %do k = 1 %to &XnVars;
	     &&XVar&k = &&AddVar&k;
	  %end;
	  %* Delete record if both geneid and intensity are missing;
	  genevtype=vtype(&gene);
	  if genevtype="N" then do;
	    if &Gene=. and &KeepI=. then delete;
		end;
	  else do;
	    if &Gene=" " and &KeepI=. then delete;
		end;
	  keep &Gene &AllVars &AllXVars &KeepI 
         %if %upcase(&Dataplatform)=QUANTARRAY %then %if ^&DelName %then Name; ;
   run;

   proc sort data=TmpDAT;
      by &Gene;
   run;

   %if %sysfunc(exist(OutLib.&OutData)) %then 
      %do;
		%if %upcase(&Dataplatform)=AGILENT  %then %do;
		   %let dsid=%sysfunc(open(TMPDAT,i));
		   %let dsid2=%sysfunc(open(OutLib.&OutData,i));
		   %let len1 = 0;
		   %do nv = 1 %to %sysfunc(attrn(&dsid,NVARS));
		      %if %sysfunc(varlen(&dsid,&nv)) > %sysfunc(varlen(&dsid2,&nv)) %then %do;
			     %let nv = %sysfunc(attrn(&dsid,NVARS));
				 %let len1 = 1;
			  %end;
		   %end;
		   %let rc=%sysfunc(close(&dsid));
		   %let rc=%sysfunc(close(&dsid2));
		   %if &len1 %then %do;
			  proc datasets library=OutLib nolist;
		         append base=Work.TmpDAT data=OutLib.&OutData;	
                 delete &OutData;
                 copy in=Work out=OutLib;
                 select TmpDAT;
                 change TmpDAT=&OutData;
         	  quit;
		   %end;
		   %else %do;
			  proc datasets library=Work nolist;
		         append base=OutLib.&OutData data=TmpDAT;
         	  quit;
	  	   %end;
		%end;
		%else %do;
	     proc datasets library=Work nolist;
		    append base=OutLib.&OutData data=TmpDAT;
         quit;
		%end;
	  %end;
   %else
      %do;
	     proc datasets library=OutLib nolist;
		    %if "%sysfunc(pathname(work))"="%sysfunc(pathname(OutLib))" %then %do;
               change TmpDAT=&OutData;
			%end;
			%else %do;
               copy in=work out=OutLib;
               select TmpDAT;
               change TmpDAT=&OutData;
			%end;
            modify &OutData (label="Data for Project: &ProjectID");
         quit;
	  %end;
%end;
%* end of parsing;

%let day2 = %sysfunc(round(%sysevalf(%sysfunc(date()))));
%let VeryEnd = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%if &day1 = &day2 %then %do;
   %let TotalTime = %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;
%else %do;
   %let TotalTime = %sysevalf((&day2-&day1)*86400) + %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;  

data _null_;
   t = &TotalTime;
   put "The total importing time is" t time11.2;
run;
%put DATParse completed.;
%mend DATParse;


****************************************************************************;
*****   Parse CDF: convert Affy's .CDF file into SAS data (ProbeMap_)  *****;
****************************************************************************;
%macro CDFParse;
data CDF (label = "Probe Map");
   retain Unit 0 TmpID '        ';
   infile "&CDFInFile" 
	      MISSOVER TRUNCOVER  
          DLM='090d'X;
   Length X X1 $17 Unit 5 X_Coordinate 3 Y_Coordinate 3 Probe_Set_ID $32 Flag $1 Probe 3;
   INPUT  X $ @;

   X1 = upcase(scan(x,1,'='));
   if X1 = 'CELLHEADER' then delete;
   if X1 = 'UNITNUMBER' then Unit = input(scan(x,2,'='),5.);
   if X1 =: '[QC' then TmpID = compress(X,'[]');
   if X1 =: 'CELL' then do;
      if unit = 0 then do; 
	     * importing QC probes;
	     input Y_Coordinate F3 $ F4 $ Probe $ F5 $ Flag $ F8 $;
		 * Flag can be '.' if the original value is '-1'
		            or ' ' if the original value is empty;
         Probe_Set_ID = TmpID;
      end;
	  else do;
         input  Y_Coordinate F3 $ F4 $ Probe_Set_ID $ F6 $ F7 $
                F8 $ PBase $ TBase $ Probe;
         if pbase=tbase then Flag = '0';
                        else Flag = '1';
	  end;
      X_Coordinate=input(scan(X,2,'='),4.);
   end;
   else delete;

   *label Unit='Unit No'
         Probe_Set_ID='Affy Internal ID'
         Probe='Probe No'
	     Flag='PM Indicator';
   label Flag = 'PM Indicator';
   Keep Unit X_Coordinate Y_Coordinate Probe_Set_ID Probe Flag;
run;

Proc sort data=CDF;
   by Y_Coordinate X_Coordinate;
run;
%put CDFParse completed.;
%mend CDFParse;


***********************************************************************;
*****   Parse CEL (probe data):                                   *****;
*****   convert Affy's .CEL files into SAS data combining with    *****;
*****   experiment factors.                                       *****;
***********************************************************************;
%macro CELParse;
%let VeryStart = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%let Day1 = %sysfunc(round(%sysevalf(%sysfunc(date()))));
 
%if %sysfunc(exist(OutLib.&OutData)) %then 
   %do;
      proc datasets library=OutLib nolist; 
         delete &OutData probemap qcprobe; 
      quit;
   %end;

%* Create probemap which contains annotation for coordinates;
%if %upcase(&ProbeUsed)=PAIR %then %do;
   %let PU= ;
   %let DropFlag= ;
%end;
%else %do;
   %let PU=%str(and Flag='1'); 
   %let DropFlag=Flag;
%end;
Proc sort data=CDF (where=(Probe_Set_ID^=:'QC' &PU))
          out=OutLib.probemap (Drop=&DropFlag);
   by &Gene Probe &DropFlag;
run;

* screen out empty file;
data ETable;
   set OutLib.exptable;
   if file ne '';
run;

%nobsvars(ETable);
%global nRawFiles RawFiles;
%let nRawFiles = &nobs;

%* start parsing;
%do i_chip = 1 %to &nRawFiles;
   data _null_;
     set ETable;
     if _n_ = &i_chip then do;
	    call symput('CEL',trim(left(File)));
        %do j = 1 %to %eval(&Fnum-1);
	       call symput("V&&j",&&VarName&j);
        %end;
	 end;
   run;
   %do j = 1 %to %eval(&Fnum-1); 
      %if &&VarType&j = N %then %do; %let tmp&j = &&V&j; %end;
	                      %else %do; %let tmp&j = "&&V&j"; %end;
   %end;
   %let flagdot = 0;
   %if %length(&CEL)>4 %then %do;
      %if %upcase(%substr(&CEL,%length(&CEL)-3,4))=.CEL %then %do;
         %let datafile=&DataInPath.&CEL;
 	     %let flagdot = 1;
	  %end;
   %end; 
   %if &flagdot=0 %then %do; 
	  %if %sysfunc(fileexist(&DataInPath.&CEL..CEL)) %then %do;
         %let datafile=&DataInPath.&CEL..CEL;
		 %let CEL = &CEL..CEL;
	  %end;
      %else %if %sysfunc(fileexist(&DataInPath.&CEL..cel)) %then %do;
         %let datafile=&DataInPath.&CEL..cel;
		 %let CEL = &CEL..cel;
	  %end;
	  %else %put ERROR: &DataInPath.&CEL does not exist.;
   %end; 
   %if %upcase(&UploadRawFiles) = YES %then %do;
      %if &i_chip = 1 %then %let RawFiles = &CEL;
      %else %let RawFiles = &RawFiles,&CEL;
   %end;

   /* import CEL files */ 
   %if &CelVersion = 3 %then %do; 
      data CEL (label = "Probe Data");
         infile "&Datafile" DLM='090d'X ;
	     Length X_Coordinate 3 Y_Coordinate 3 Mean 4;
         %*** Find the beginning of the Intensity Section;
         input @'[INTENSITY]' ; 
         %*** Find the key word for Number of Cells and obtain value;
         input @'NumberCells=' numCells :9.;
         %*** Find data header line (data begins on next line);
         input @'CellHeader=';
      
	     do i = 1 to numCells;
            INPUT  X_Coordinate Y_Coordinate Mean ;
            %* Copy the values of experimental factors;
            %do j = 1 %to %eval(&Fnum-1); 
               &&VarName&j = &&tmp&j;
            %end;  
            output;
         end; 
         drop numCells i;
      run;
   %end;
   %else %do;
      data CEL (label = "Probe Data");
         infile "&Datafile" RECFM=N lrecl=32767;
         length X_Coordinate 3 Y_Coordinate 3 Mean 4;
         if _n_=1 then do;
            input +8 nCol IBR4. +4 nRec IBR4. nSkip IBR4. +nSkip nSkip IBR4.
                  +nSkip nSkip IBR4. +nSkip +16 @@;
            retain nCol nRec;
         end;
         X_Coordinate = _n_ - (ceil(_n_/nCol)-1)*nCol - 1;
         Y_Coordinate = ceil(_n_/nCol) - 1;
         input Mean IEEER4. STDV IEEER4. nPixels IBR2. @@;
         %* Copy the values of experimental factors;
         %do j = 1 %to %eval(&Fnum-1); 
            &&VarName&j = &&tmp&j;
         %end;  
         if _n_ > nRec then stop;
		 drop nCol nRec nSkip STDV nPixels;
      run;
   %end;

   %if &i_chip = 1 %then %do;
      %nObsVars(CEL);
      %if &nobs=0 %then %do;
         %put ERROR: Error(s) were encountered importing CEL file. This error may be caused by specifying incorrect DataPlatform;
         %let exiterror = 1;
		 %goto exit;
      %end;
   %end;

   proc sort data=CEL;
      by Y_Coordinate X_Coordinate;
   run;

   %let PU = ;
   %let DropFlag2 = ;
   %let KeepFlag = Flag;
   %if %upcase(&ProbeUsed) = PM %then %do;
      %let PU = %str(and (Flag='1' or Probe_Set_ID =: 'QC')); %*keep PM only;
      %let DropFlag2 = Flag;
	  %let KeepFlag = ;
   %end;

   data CEL(drop=Y_Coordinate X_Coordinate mean &DropFlag2)
        QC(drop=mean);
      merge CDF(in=CDF) CEL(in=CEL);
      by Y_Coordinate X_Coordinate;
	  if CDF and CEL &PU;
	  %if %upcase(&Log2Transform) = YES %then %do;
         if Mean>0 then log2i = log2(Mean);
	               else log2i = .;
	  %end;
	  %else %do;
         Intensity = Mean;
	  %end;
	  if Probe_Set_ID =: 'QC' then output QC;
                        else output CEL; 
   run;

   proc sort data=CEL;
      by &Gene Probe &KeepFlag;
   run;

   %if %upcase(&ProbeUsed) = DIFFERENCE %then %do;
      %if %upcase(&Log2Transform) = YES %then %let Iname = log2i;
	                                    %else %let Iname = Intensity;

      data CEL;  
	     merge CEL(where=(Flag='0') rename=(&Iname=&Iname._m))
               CEL(where=(Flag='1'));
		 by &Gene Probe;
		 %if %upcase(&Log2Transform) = YES %then %do;
            if log2i-log2i_m > 0 then log2d = log2(2**log2i-2**log2i_m);
	                             else log2d = .;
			Label log2d="log2(PM-MM)";
		    drop flag log2i log2i_m;
		 %end;
		 %else %do;
		    PM_MM = Intensity - Intensity_m; 
            Label PM_MM="PM - MM";
			drop flag Intensity Intensity_m;
		 %end;
	  run;
   %end;

   %if %sysfunc(exist(OutLib.&OutData)) %then 
      %do;
	     proc datasets library=Work nolist;
		    append base=OutLib.&OutData data=CEL;
		    append base=OutLib.qcprobe data=QC;
         quit;
	  %end;
   %else
      %do;
	     proc datasets library=OutLib nolist;
		    %if "%sysfunc(pathname(work))"="%sysfunc(pathname(OutLib))" %then %do;
               change CEL=&OutData;
			   change QC=qcprobe;
			%end;
			%else %do;
               copy in=work out=OutLib;
               select CEL QC;
               change CEL=&OutData QC=qcprobe;
			%end;
            modify &OutData (label="CELs for Project: &ProjectID");
            modify qcprobe (label="QC probes for Project: &ProjectID");
         quit;
	  %end;
%end;
%* end of parsing;

%let day2 = %sysfunc(round(%sysevalf(%sysfunc(date()))));
%let VeryEnd = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%if &day1 = &day2 %then %do;
   %let TotalTime = %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;
%else %do;
   %let TotalTime = %sysevalf((&day2-&day1)*86400) + %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;  

data _null_;
   t = &TotalTime;
   put "The total importing time is" t time11.2;
run;
%put CELParse completed.;

%exit:;
%mend CELParse;


***********************************************************************;
*****   Parse CHP (signal data):                                  *****;
*****   convert Affy's .CHP files into SAS data combining with    *****;
*****   experiment factors.                                       *****;
***********************************************************************;
%macro CHPParse;
%let VeryStart = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%let Day1 = %sysfunc(round(%sysevalf(%sysfunc(date()))));
 
* screen out empty file;
data ETable;
   set OutLib.exptable;
   if file ne '';
run;

%if %sysfunc(exist(OutLib.&OutData)) %then 
   %do;
      proc datasets library=OutLib nolist; 
         delete &OutData; 
      quit;
   %end;

%nobsvars(ETable);
%global nRawFiles RawFiles;
%let nRawFiles = &nobs;

%* start parsing;
%do i_chip = 1 %to &nRawFiles;
   data _null_;
     set ETable;
     if _n_ = &i_chip then do;
	    call symput('CHP',trim(left(File)));
        %do j = 1 %to %eval(&Fnum-1);
	       call symput("V&&j",&&VarName&j);
        %end;
	 end;
   run;
   %do j = 1 %to %eval(&Fnum-1); 
      %if &&VarType&j = N %then %do; %let tmp&j = &&V&j; %end;
	                      %else %do; %let tmp&j = "&&V&j"; %end;
   %end;
   %let flagdot = 0;
   %if %length(&CHP)>4 %then %do;
      %if %upcase(%substr(&CHP,%length(&CHP)-3,4))=.CHP %then %do;
         %let datafile=&DataInPath.&CHP;
 	     %let flagdot = 1;
	  %end;
   %end; 
   %if &flagdot=0 %then %do; 
	  %if %sysfunc(fileexist(&DataInPath.&CHP..CHP)) %then %do;
         %let datafile=&DataInPath.&CHP..CHP;
		 %let CHP = &CHP..CHP;
	  %end;
      %else %if %sysfunc(fileexist(&DataInPath.&CHP..chp)) %then %do;
         %let datafile=&DataInPath.&CHP..chp;
		 %let CHP = &CHP..chp;
	  %end;
	  %else %put ERROR: &DataInPath.&CHP does not exist.;
   %end; 
   %if %upcase(&UploadRawFiles) = YES %then %do;
      %if &i_chip = 1 %then %let RawFiles = &CHP;
      %else %let RawFiles = &RawFiles,&CHP;
   %end;

   /* import CHP files */ 
   data CHP (label = "Signal Data");
      infile "&datafile" RECFM=N lrecl=32767;
      length X_Coordinate 3 Y_Coordinate 3 Signal 4; 
      retain nUnit count flag Version 0;
      count = count+1;
      if _n_=1 then do;
         input +22 Version IBR4. nSkip IBR4. +nSkip nSkip IBR4. +nSkip nSkip IBR4.
               +nSkip nSkip IBR4. +nSkip +8 nUnit IBR4. nMax IBR4.
               +(nUnit*8) +(nMax*8) +516 nSkip IBR4. +nSkip @;
      end;
      if Version=12 then input nProbe IBR4. +32 Signal IEEER4. +12 X_Coordinate IBR4.  
                               Y_Coordinate IBR4. +(52*nProbe-16) flag IBR4. @;
                    else input nProbe IBR4. +12 Signal IEEER4. +12 X_Coordinate IBR2. 
                               Y_Coordinate IBR2. +(16*nProbe-12) flag IBR4. @;
      if flag then do;
         if Version = 12 then input +58 @;
         else input +29 @; 
      end; 
      %* Copy the values of experimental factors;
      %do j = 1 %to %eval(&Fnum-1); 
         &&VarName&j = &&tmp&j;
      %end;  
      if count>nUnit then stop;
	  drop Version nSkip nUnit nMax nProbe flag count;
   run;

   %if &i_chip = 1 %then %do;
      %nObsVars(CHP);
      %if &nobs=0 %then %do;
         %put ERROR: Error(s) were encountered importing CHP file. This error may be caused by specifying incorrect DataPlatform;
         %let exiterror = 1;
		 %goto exit;
      %end;
   %end;

   proc sort data=CHP;
      by Y_Coordinate X_Coordinate;
   run;

   data CHP;
      merge CDF(in=CDF) CHP(in=CHP);
      by Y_Coordinate X_Coordinate;
	  if CDF and CHP;
	  %if %upcase(&Log2Transform) = YES %then %do;
         if Signal>0 then log2i = log2(Signal);
	                 else log2i = .;
	  %end;
	  %else %do;
         Intensity = Signal;
	  %end;
	  drop Y_Coordinate X_Coordinate Signal Flag Probe;
   run;

   proc sort data=CHP;
      by &Gene;
   run;

   %if %sysfunc(exist(OutLib.&OutData)) %then 
      %do;
	     proc datasets library=Work nolist;
		    append base=OutLib.&OutData data=CHP;
         quit;
	  %end;
   %else
      %do;
	     proc datasets library=OutLib nolist;
		    %if "%sysfunc(pathname(work))"="%sysfunc(pathname(OutLib))" %then %do;
               change CHP=&OutData;
			%end;
			%else %do;
               copy in=work out=OutLib;
               select CHP;
               change CHP=&OutData;
			%end;
            modify &OutData (label="CHPs for Project: &ProjectID");
         quit;
	  %end;
%end;
%* end of parsing;

%let day2 = %sysfunc(round(%sysevalf(%sysfunc(date()))));
%let VeryEnd = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%if &day1 = &day2 %then %do;
   %let TotalTime = %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;
%else %do;
   %let TotalTime = %sysevalf((&day2-&day1)*86400) + %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;  

data _null_;
   t = &TotalTime;
   put "The total importing time is" t time11.2;
run;
%put CHPParse completed.;

%exit:;
%mend CHPParse;


**********************************************************************;
*****   Parse Table: convert table formated file into SAS data   *****;
**********************************************************************;
%macro TableParse;
%let VeryStart = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%let Day1 = %sysfunc(round(%sysevalf(%sysfunc(date()))));

data ETable;
   set OutLib.exptable;
   if file ne '';
run;

%* start parsing;
%if %sysfunc(exist(OutLib.&OutData)) %then 
   %do;
      proc datasets library=OutLib nolist; 
         delete &OutData; 
      quit;
   %end;

%global nRawFiles RawFiles;
data ETable;
   set ETable;
   n = _n_;
run;
proc sort data=ETable(keep=File n) out=tmp nodupkey;
   by File;
run;
proc sort data=tmp(drop=file);
   by n;
run;
data ETable;
   merge ETable tmp(in=in);
   by n;
   if in then Flag = 1;
   drop n;
run;
%nobsvars(tmp);
%let nRawFiles = &nObs;
%let nRec = 0;

%nobsvars(ETable);
%let nRows = &nObs;
%let nCols = %eval(&nVars-1);
%if %upcase(&Log2Transform) = YES %then %let KeepI = Log2i;
                                  %else %let KeepI = Intensity;
***** Added optional delimiter statement to handle bogus end-of-line characters;
%if %upcase(&DataType) = CSV %then %let DLT = DELIMITER='2C0D'x;
%else %if %upcase(&DataType) = DLM %then %let DLT = DELIMITER='200D'x;
%else %if %upcase(&DataType) = TAB %then %let DLT = DELIMITER='090D'x;
%else %let delimiter = ;

%do iArray = 1 %to &nRows; 
   %let FileFlag = 0;
   data _null_;
      set ETable;
      if _n_ = &iArray;
	  call symput('Tab',trim(left(File)));
	  call symput('Int',trim(left(Intensity)));
      %do j = 1 %to %eval(&nCols-&XnVars-1);
	     call symput("V&j",&&VarName&j);
	     call symput("T&j","&&VarType&j");
      %end;
	  %do k = 1 %to &XnVars;
	     call symput("AddVar&k",&&XVar&k);
	  %end;
	  if Flag then call symput("FileFlag",1);
   run;

%if &FileFlag %then %do; * avoid duplicate importing;
   /* import table formated files */
   %if %upcase(&DataType) = SAS %then %do;
      %let L = %eval(%index(%upcase(&Tab),.SAS7BDAT)-1);
      %let Tab = %substr(&Tab,1,&L);
	  libname TmpLib "&DataInPath";
	  proc datasets library=work nolist;
		    %if "%sysfunc(pathname(work))"="%sysfunc(pathname(TmpLib))" %then %do;
               %if %upcase(&Tab) ne TABLE %then change &Tab=Table; ;
			%end;
			%else %do;
			   %if %sysfunc(exist(Table)) %then delete Table; ;
               copy in=TmpLib out=Work;
               select &Tab;
               change &Tab=Table;
			%end;
      quit;
   %end;
   %else %do; 
      %let DataFile = &DataInPath.&Tab; 
      %if &iArray = 1 %then %do;
	     %let VarList = ;
		 %let LenList = ;
	     %skip_header_import_2(DataFile=%quote(&DataFile),DataType=&DataType,
            OutSet=Table,FirstRow=&DataStartRow,VarNameRow=&VarNameRow,
            GuessRows=&ScanRows,_varlist=VarList,_lenlist=LenList);
		 %if %length(&DataType)>3 & &VarNameRow>0 %then %do; 
		    %let tid = %sysfunc(open(Table));
            %let nv_Table = %sysfunc(attrn(&tid,NVARS)); 
            %do i = 1 %to &nv_Table;
               %let _name&i = %sysfunc(VARNAME(&tid,&i));
            %end;
            %let tid = %sysfunc(close(&tid));
		 %end;
	  %end;
	  %else %do;
	     * Excel files;
	     %if %length(&DataType)>3 %then %do;
            proc import out=table 
                 datafile= "&DataFile"
                 dbms=&DataType replace;
               Getnames=no;
		       %if &DataStartRow>1 %then range="A&DataStartRow.:IW100000000";;
			run;
            %if &VarNameRow>0 %then %do;
               proc datasets lib=work nolist;
		          modify table;	     
                  %do i = 1 %to &nv_Table;
	                 %if F&i ne &&_name&i %then rename F&i = &&_name&i;;
	              %end; 
	           quit;
			%end;
		 %end;
		 %else %do;
            data Table;
               infile "&datafile" &DLT MISSOVER DSD lrecl=32767 firstobs=&DataStartRow;
               length &LenList;
               input &VarList;
            run;
		 %end;
	  %end;
   %end; 

   %if %upcase(&UploadRawFiles) = YES %then %do;
      %let nRec = %eval(&nRec+1);
      %if &nRec = 1 %then  %let RawFiles = &Tab;
      %else %let RawFiles = &RawFiles,&Tab;
   %end;
%end;

   data TmpTable;
      set Table;
	  %* copy the values of experimental factors;
	  %do j = 1 %to %eval(&nCols-&XnVars-1);
	     %if &&T&j = N %then %do; &&VarName&j = &&V&j; %end;
	                   %else %do; &&VarName&j = "&&V&j"; %end;
	  %end;
	  %if %upcase(&Log2Transform) = YES %then %do;
	     if &Int>0 then log2i = log2(&Int);
	               else log2i = .;
	  %end;
      %else %do;
	     Intensity = &Int;
	  %end;
	  %do k = 1 %to &XnVars;
	     &&XVar&k = &&AddVar&k;
	  %end;
	  %* Delete record if both first id var and intensity are missing;
	  if vtype(&ID1)="N" then do;
	    if &ID1=. and &KeepI=. then delete;
		end;
	  else do;
	    if &ID1=" " and &KeepI=. then delete;
		end;
	  keep &IDVars &AllVars &AllXVars &KeepI;
   run;

   proc sort data=TmpTable;
      by &IDVars;
   run;
   
   * check if there is empty intensity on top which is
     caused by extra empty lines in the raw data;
   %if %upcase(&Log2Transform) = YES %then %let IVName = log2i;
                                     %else %let IVName = Intensity;
   %let nMiss = 0;
   data _null_;
      set TmpTable;
	  if &IVName ^= . then do;
	     call symput('nMiss',_n_);
         stop;
      end;
   run;
   %if &nMiss %then %do;
      data TmpTable;
         set TmpTable;
	     if _n_ < &nMiss then delete;
      run;
   %end;

   %if %sysfunc(exist(OutLib.&OutData)) %then 
      %do;
	     proc datasets library=Work nolist;
		    append base=OutLib.&OutData data=TmpTable;
         quit;
	  %end;
   %else
      %do;
	     proc datasets library=OutLib nolist;
		    %if "%sysfunc(pathname(work))"="%sysfunc(pathname(OutLib))" %then %do;
               change TmpTable=&OutData;
			%end;
			%else %do;
               copy in=work out=OutLib;
               select TmpTable;
               change TmpTable=&OutData;
			%end;
            modify &OutData (label="Data for Project: &ProjectID");
         quit;
	  %end;
%end;
%* end of parsing;

%let day2 = %sysfunc(round(%sysevalf(%sysfunc(date()))));
%let VeryEnd = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%if &day1 = &day2 %then %do;
   %let TotalTime = %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;
%else %do;
   %let TotalTime = %sysevalf((&day2-&day1)*86400) + %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;  

data _null_;
   t = &TotalTime;
   put "The total importing time is" t time11.2;
run;
%mend TableParse;
 


%macro skip_header_import(DataFile=,OutData=Dat,VarLen=1000,FirstRow=,VarNameRow=);
*Retrieve variable names;
DATA TmpData;
   INFILE "&DATAFILE" MISSOVER DSD lrecl=32767 firstobs=&VarNameRow OBS=%eval(&firstrow+&scanrows);
   LENGTH var1 $ &varlen;
   input var1;
   if _n_ = 1 | _n_ > %eval(&FirstRow-&VarNameRow);
run;

%let workpath=%sysfunc(pathname(work));
%if &sysscp=WIN %then %let _slash=\;
%else %let _slash=/;

*put the variable name and 20 rows of data in a temp file;
Data _null_;
   set TmpData;
   file "&workpath.&_slash.import_temp.txt" lrecl=32767;
   put var1;
run;

*Import the temp file to get the format;
PROC IMPORT OUT=import_try DATAFILE= "&workPath.&_slash.import_temp.txt" 
      DBMS=DLM REPLACE;
   DELIMITER='090d'x; 
   GETNAMES=YES;
   DATAROW=2; 
RUN;

proc contents data=import_try out=datype(keep=NAME TYPE VarNum length) noprint;
run;

proc sort data=datype;
   by varnum;
run;

*Get the input variable list;
%let logrt=0;
Data _null_;
   set datype end=eof;
   length inputvar lenvar $&VarLen;
   retain inputvar lenvar;
   %if %upcase(&DataPlatform) = GENEPIX %then %do;
      if type=1 and substr(left(upcase(name)),1,9) ne "LOG_RATIO" then 
         inputvar=left(trim(inputvar))||" "||name;
      *for log ratio variable, set it to character and then convert to 
	   numeric to avoid the error message.  This adjustment applys for 
	   GenePix data.;
      else if type=1 and substr(left(upcase(name)),1,9) = "LOG_RATIO" then do;
	 	 inputvar=left(trim(inputvar))||" log_r $";
		 lenvar=left(trim(lenvar))||" log_r $8";
		 call symput('logrvar',name);
		 call symput('logrt','1');
      end;
   %end;
   %else %do;
      if type=1 then inputvar=left(trim(inputvar))||" "||name;
   %end;
   else if type=2 then do;
      inputvar=left(trim(inputvar))||" "||trim(name)||" "||"$";
      lenvar=left(trim(lenvar))||" "||trim(name)||" "||"$"||trim(left(put(length,3.)));
   end;
   if eof then do; 
      call symput('inputvar',trim(left(inputvar)));
	  call symput('lenvar',trim(left(lenvar)));
   end;
run;

*Read in the final data;
options errors=1;
Data &outdata;
   infile "&datafile" delimiter='090d'x MISSOVER DSD lrecl=32767 firstobs=&firstrow;
   length &lenvar;
   input &inputvar;
   %if &logrt %then %do; 
      if upcase(log_r) ^in("ERROR","NA","N/A") then &logrvar=log_r+0;
      drop log_r;
   %end;   
run;

options errors=30;
%if &syserr %then %do;
   %put ;
   %put "Note: When you see an error above, it may be due to some characters read as numerical data.";
   %put "It can be ignored in this case.";
   %put ;
%end;
%mend skip_header_import;


%macro skip_header_import_2(DataFile=,DataType=,OutSet=Dat,VarLen=32767,FirstRow=,
                            VarNameRow=0,GuessRows=20,_VarList=,_LenList=);
%if &sysscp=WIN %then %let _slash=\;
%else %let _slash=/;
***** Added optional delimiter statement to handle bogus end-of-line characters;
%if %upcase(&DataType) = CSV %then %let DLT = delimiter='2C0D'x;
%else %if %upcase(&DataType) = DLM %then %let DLT = delimiter='200D'x;
%else %if %upcase(&DataType) = TAB %then %let DLT = delimiter='090D'x;
%else %let DLT = ; 

%if ^%length(&_VarList) %then %let _VarList = _TmpList1;
%if ^%length(&_LenList) %then %let _LenList = _TmpList2;

*Retrieve variable names;
%if &VarNameRow %then %do; 
   %if %upcase(%substr(&DataType,1,%sysfunc(min(5,%length(&DataType)))))=EXCEL %then %do;
	  * adjust naming convention for Excel file;
      proc import out=tmpe 
           datafile= "&DataFile"
           dbms=&DataType replace;
         Getnames=No;
	     range="A&VarNameRow.:IW&VarNameRow";
      run;
	  %nObsVars(tmpe); 
	  %let Vars = ;
	  %do i = 1 %to &nVars;
	     %let Vars = &Vars F&i;
	  %end;	  
      %naming(Dset=tmpe,nVars=&nVars,AllVars=&Vars);
	  data _null_;
	     set tmpe;
	     %do i = 1 %to &nVars;
	         call symput("TmpName&i",F&i);
	    %end;
	  run;
   %end;
   %else %do;
      * adjust naming convention for text file;
      %let workpath=%sysfunc(pathname(work));
      filename tmpfile "&workpath.&_slash.tmpfile072304.txt";
      data _null_;
         infile "&DataFile" truncover lrecl=&VarLen;
	     length x $&VarLen;
	     input x 1-&VarLen;
	     if _n_=&VarNameRow then do;
	        file tmpfile lrecl=&VarLen;
	        put x;
	     end;
	     if _n_ > &VarNameRow then stop;
	  run;
      proc import out=tmpe 
           datafile= "&workpath.&_slash.tmpfile072304.txt"
           dbms=&DataType replace;
         Getnames=No;
	     DataRow=1;
      run; 
      %if %sysfunc(fexist(tmpfile)) %then %let frc=%sysfunc(fdelete(tmpfile));
	  %nObsVars(tmpe); 
      %let Vars = ;
	  %do i = 1 %to &nVars;
	     %let Vars = &Vars Var&i;
	  %end;	
      %naming(Dset=tmpe,nVars=&nVars,AllVars=&Vars);
	  data _null_;
	     set tmpe;
	     %do i = 1 %to &nVars;
	        call symput("TmpName&i",Var&i);
	     %end;
	  run;
   %end;
%end;

* Read full data;
%if %length(&DataType)>3 %then %do;
   PROC IMPORT OUT= &OutSet 
        DATAFILE= "&DataFile"
        DBMS=&DataType REPLACE;
      GETNAMES=no; 
      %if &FirstRow>1 %then %do;
         range="A&FirstRow.:IW100000000";
      %end;
   RUN;
   %if &VarNameRow %then %do;
      %let tid = %sysfunc(open(&OutSet));
      %let nv_OutSet = %sysfunc(attrn(&tid,NVARS)); 
      %do i = 1 %to &nv_OutSet;
         %let _name&i = %sysfunc(VARNAME(&tid,&i));
	  %end;
      %let tid = %sysfunc(close(&tid));
   %end;
%end;
%else %do;
   %let workpath=%sysfunc(pathname(work));
   filename tmpfile "&workpath.&_slash.tmpfile072204.txt";
   * output a partial text file contianing data only for determining variable attributes;
   data _null_;
      infile "&DataFile" truncover lrecl=&VarLen;
	  length x $&VarLen;
	  input x 1-&VarLen;
	  if _n_ > %eval(&FirstRow+&GuessRows-1) then stop;
	  if _n_>=&FirstRow then do;
	     file tmpfile lrecl=&VarLen;
	     put x;
	  end;
   run;
   * Import data again for getting variable attributes;
   proc import out=&OutSet
        datafile= "&workpath.&_slash.tmpfile072204.txt"
        dbms=&DataType replace;
      GETNAMES=no; 
	  DataRow = 1; 
   run; 
   %if %sysfunc(fexist(tmpfile)) %then %let frc=%sysfunc(fdelete(tmpfile));
   * Read data final time;
   %let &_VarList = ;
   %let &_LenList = ;
   %let tid = %sysfunc(open(&OutSet));
   %let nv_OutSet = %sysfunc(attrn(&tid,NVARS)); 
   %do i = 1 %to &nv_OutSet;
      %let _name&i = %sysfunc(VARNAME(&tid,&i));
      %let _len&i = %sysfunc(VARLEN(&tid,&i));
      %let _type&i = %sysfunc(VARTYPE(&tid,&i));
	  %if &&_type&i = N %then %do;
	     %let &_LenList = &&&_LenList &&_name&i &&_len&i...;
	     %let &_VarList = &&&_VarList &&_name&i;
	  %end;
	  %else %do;
	     %let &_LenList = &&&_LenList &&_name&i $&&_len&i...;
	     %let &_VarList = &&&_VarList &&_name&i $;
	  %end;
   %end;
   %let tid = %sysfunc(close(&tid));
   * apply data step to control variable attributes;
   data &OutSet;
      infile "&datafile" &DLT MISSOVER DSD lrecl=&VarLen firstobs=&firstrow;
      length &&&_LenList;
      input &&&_VarList;
   run;
%end;

* renmae variables;
%if &VarNameRow %then %do;
   %let nMax = %sysfunc(min(&nv_OutSet,&nVars));
   %if %scan(&OutSet,1) = &OutSet %then %do;
      %let _lib = work;
      %let _dset = &OutSet;
   %end;
   %else %do;
	  %let _lib = %scan(&OutSet,1);
	  %let _dset = %scan(&OutSet,2);
   %end; 
   proc datasets lib=&_lib nolist;
	  modify &_dset;	     
      %do i = 1 %to &nMax;
	     %if &&_name&i ne &&TmpName&i %then rename &&_name&i = &&TmpName&i;;
	  %end; 
   quit;
%end;
* make variable name and length list;
%let &_VarList = ;
%let &_LenList = ;
%let tid = %sysfunc(open(&OutSet));
%let nv_OutSet = %sysfunc(attrn(&tid,NVARS)); 
%do i = 1 %to &nv_OutSet;
   %let _name&i = %sysfunc(VARNAME(&tid,&i));
   %let _len&i = %sysfunc(VARLEN(&tid,&i));
   %let _type&i = %sysfunc(VARTYPE(&tid,&i));
   %if &&_type&i = N %then %do;
      %let &_LenList = &&&_LenList &&_name&i &&_len&i...;
      %let &_VarList = &&&_VarList &&_name&i;
   %end;
   %else %do;
      %let &_LenList = &&&_LenList &&_name&i $&&_len&i...;
      %let &_VarList = &&&_VarList &&_name&i $;
   %end;
%end;
%let tid = %sysfunc(close(&tid));
%mend skip_header_import_2;

%macro CheckCoorVar(_ds=,_Old=,_New=);
%let tid = %sysfunc(open(&_ds));
%let nv = %sysfunc(attrn(&tid,NVARS));
%let &_New = ;
%let i = 1;
%do %while(%length(%scan(&_Old,&i)));
   %do j = 1 %to &nv;
      %if %upcase(%sysfunc(VARNAME(&tid,&j)))=%upcase(%scan(&_Old,&i)) %then
         %let &_New = &&&_New %scan(&_Old,&i);
   %end;
   %let i = %eval(&i+1);
%end;
%let tid = %sysfunc(close(&tid));
%mend CheckCoorVar;


%macro CheckInputs;
%let clockrunning = 1;
%ClockStart(&ProcessName);
%if &syserr %then %goto exit;
%if %length(&ProjectID)=0 %then %let ProjectID=&ProcessName;
%CheckForBlank(&DataInPath,Full path of data);
%if &exiterror %then %goto exit;
%CheckForBlank(&ExpInFile,Experimental design table);
%if &exiterror %then %goto exit;
%CheckForBlank(&ExpType,Data type of Experimental design table);
%if &exiterror %then %goto exit;
%if %length(&OutPath)=0 %then %let OutPath=%sysfunc(pathname(work));

%exit:;
%if (&syserr | &exiterror) %then %put ERROR: Error(s) occur during checking inputs.;
%mend CheckInputs;


%macro BeforeDataParse;
* Check if OutData is valid SAS data set name;
%checkdataname(_name=OutData,_default=&ProjectID);

* create output libaray;
libname OutLib "&OutPath";

%ImportScanRows(Flag=1,ScanRows=&ScanRows);
%DesignParse;
%mend BeforeDataParse;


%macro AfterDataParse;
%NObsVars(OutLib.&OutData);
%if %sysfunc(exist(OutLib.exitedwitherrors)) %then %do;
   proc datasets lib=OutLib nolist; 
      delete exitedwitherrors; 
   run;
%end;
%if (&nobs=0 | &nvars=0) %then %do;
   data OutLib.exitedwitherrors;
      error=1;
   run;
   %put ERROR: Error(s) were encountered creating output combining data set.;
   %let exiterror = 1;
%end;
%ImportScanRows(Flag=0);
%mend AfterDataParse;


%macro Packaging;
%PackageBegin(&ProcessName);
%PackageInsertDataset(data from &ProcessName,outlib,ExpTable,No);
%PackageInsertDataset(data from &ProcessName,outlib,&OutData,No);
%if %upcase(%substr(&DataPlatform,1,4))=AFFY %then %do;
   %if %upcase(%substr(&DataPlatform,6,3)) = CEL %then %do;
      %NObsVars(OutLib.qcprobe);
      %if (&nobs=0 | &nvars=0) %then 
	     %put WARNING: Data set qcprobe is empty and will not be packaged.;
	  %else 
         %PackageInsertDataset(data from &ProcessName,outlib,qcprobe,No);
      %PackageInsertDataset(data from &ProcessName,outlib,probemap,No);
   %end;
   %if %upcase(&UploadRawFiles)=YES %then %do;
      %if %upcase(%substr(&DataPlatform,6,13)) = CEL VERSION 3 %then %let mType = text/plain;
      %else %let mType = application/binary;
	  %do k = 1 %to &nRawFiles;
	     %let TmpFile = %scan(%quote(&RawFiles),&k,%quote(,));
         %PackageInsertFile(Row Data,&DataInPath,&TmpFile,&OutData..sas7bdat,&mType,No);
	  %end;
   %end;
%end;
%else %do;
   %if %upcase(&DataPlatform)=AGILENT | 
       %upcase(&DataPlatform)=GENEPIX |
       %upcase(&DataPlatform)=QUANTARRAY |
       %upcase(&DataPlatform)=SCANANALYZE %then %do;
      %PackageInsertDataset(data from &ProcessName,outlib,spotmap,No);
   %end;
   %if %upcase(&UploadRawFiles)=YES %then %do;
      %let mType = text/plain;
	  %do k = 1 %to &nRawFiles;
	     %let TmpFile = %scan(%quote(&RawFiles),&k,%quote(,));
         %PackageInsertFile(Row Data,&DataInPath,&TmpFile,&OutData..sas7bdat,&mType,No);
	  %end;
   %end;
%end;
%if %upcase(&KeepProcessInfo) = YES %then %do;
   %PackageInsertDataset(process information of &ProcessName,outlib,_processinfo,No);
%end;
%PackagePublish(&OutPath,to_requester);
%PackageEnd;
%mend Packaging;