/*
SAS Microarray ScanAlyze Input Engine

Name: ScanAlyzeInputEngine
Author: Tzu-Ming Chu, SAS Institute Inc.
Support: Tzu-Ming Chu, SAS Institute Inc.
History:
   tzuchu 28Aug02 initial coding   
   tzuchu 05Mar04 separate from Previous Input Engine 

Description: ScanAlyzeInputEngine imports user 
provided raw file from ScanAlyze output into 
SAS data set. Additional output "spotmap" 
containing spot orientation will be created.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

OPTIONS nospool;
* program start;
* define input parameters;
%let DataInPath = ; 
* path of data;

%let DataStartRow = 9;
* The row in the raw data files to begin reading data.;

%let UploadRawFiles = No; 
* Specify "Yes" to upload Raw Files (in original format) into warehouse;

%let ExpInFile = ;
* experimental design table;

%let ExpType = ; 
* file type of experimental design table. Excel Type is not accepted for 
Unix Server.;

%let Log2Transform = Yes; 
* option to take logarithm based 2 transformation on raw data.;

%let ScanRows = 20; 
* The first number of rows in each raw data file to be scanned to determinte the
attributes of variables (columns) in the combined SAS data set.  The attributes 
include type (such as numerical, charcater, and date) and length of variables. 
Notice that once the attributes are decided, it will apply to entire data file. 
The length of character variable is main concern among attributes.  If there is
any row after the first "ScanRows" having larger length than the determined length
of variable, it will be truncated.; 

%let OutPath = ; 
* Path of working library;
 
%let OutData = ; 
* SAS data for further analysis. Length greater than 32 wil be truncated. Specified
ProjectID will be applied as the name of output data when leaving this field empty.;

%let MacroPath = ;
* path of SAS macros;

%let KeepProcessInfo = No; 
* option to store process information;


*ProcessBody;

%macro ScanAlyzeIE;
* project identifier;
%let ProjectID = ; 

*initialization;
%let ProcessName = ScanAlyzeInputEngine;

%global exiterror;
%let exiterror = 0;
%let clockrunning = 0;

* invoke macros;
data _null_;
   infile "&MacroPath.UtilityMacros.sas";
run;
%if &syserr %then %do;
   %put ERROR: Path for utility macros is invalid.;
   %goto exit;
%end;
%include "&MacroPath.UtilityMacros.sas" / nosource;

data _null_;
   infile "&MacroPath.InputEngineMacros.sas";
run;
%if &syserr %then %do;
   %put ERROR: Path for input engine macros is invalid.;
   %goto exit;
%end;
%include "&MacroPath.InputEngineMacros.sas" / nosource;

%let Dataplatform = SCANALYZE;
%let Gene = Spot;
%if ^%length(&DataStartRow) %then %let DataStartRow = 9;

%CheckInputs;
%if &exiterror %then %goto exit;
%BeforeDataParse;
%if &exiterror %then %goto exit;
%DATParse;
%AfterDataParse;
%if &exiterror %then %goto exit;
%Packaging;

%exit:;
%if (&syserr | &exiterror) %then %put ERROR: &ProcessName exited due to errors.;
%if (&clockrunning) %then %ClockStop(&ProcessName);
%ImportScanRows(Flag=0);

%mend ScanAlyzeIE;
%ScanAlyzeIE;