/*
SAS Scientific Discovery Solutions 

Name: WebSearchJSL
Author: Geoffrey Mann, SAS Institute Inc.
Support: Jun Zhang, SAS Institute Inc.
History:
   gemann 28Aug02 intial coding
	JZhang 28May03 Modified the code to add a button on the web search dialog for 
						searching Onto-Express
	sasrdw 02Jul03 added annotation processing

Description: WebSearchJSL creates JSL code using PUT statements.  It assigns labels and 
generates a bioinformatics Web Search dialog in JMP using the global variables 
AnnoInFile, AnnoLabel, AnnoData, AnnoAccession, and AnnoMergeVar. It is designed to be 
%included within analytical process SAS code that is generating JSL in the same fashion.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

%macro AnnoLabelJSL;
   %let AnnoLabVar = ;
   %if %length(&AnnoLabel) %then %let AnnoLabVar = &AnnoLabel;
   %else %if %length(&AnnoMergeVar) %then %let AnnoLabVar = &AnnoMergeVar;
   %if %length(&AnnoLabVar) %then %do;
      %* check if &AnnoLabVar has label;
      %let dsid = %sysfunc(open(AnnoLib.&AnnoData));
      %let ALabel = %sysfunc(varlabel(&dsid,%sysfunc(varnum(&dsid,&AnnoLabVar))));
      %let rc1 = %sysfunc(close(&dsid));
      %if &ALabel = %then %do;
         put "%str(column(%"&AnnoLabVar%") << label(1);)";
      %end;
      %else %do;
         put "%str(column(%"&ALabel%") << label(1);)"; 
      %end;
   %end;
   put " ";
%mend AnnoLabelJSL;

%macro WebSearchJSL;

   %if %length(&AnnoAccession) %then %do;
      %* check if &AnnoAccession has label;
      %let dsid = %sysfunc(open(AnnoLib.&AnnoData));
		%let AAcc = %sysfunc(varlabel(&dsid,%sysfunc(varnum(&dsid,&AnnoAccession))));
      %let rc1 = %sysfunc(close(&dsid));
		%if %length(&AAcc) %then %let AnnoAccession = &AAcc;
		
   put 'ug = expr( For Each Row( If (Selected(), (';
   put 'selRow = Row();';
   put 'url = "http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?db=unigene&cmd=search&term="';
   put "|| :&AnnoAccession[selrow];";
   put 'web(url);wait(3);';
   put '));););';
   put " ";

   put 'GBN = expr( For Each Row( If (Selected(), (';
   put 'selRow = Row();';
   put 'url = "http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?db=nucleotide&cmd=search&term="';
   put "|| :&AnnoAccession[selrow];";
   put 'web(url);wait(3);';
   put '));););';
   put " ";

   put 'flyBaseSearch = expr( For Each Row( If (Selected(), (';
   put 'selRow = Row();';
   put 'url = "http://flybase.bio.indiana.edu/.bin/fbgenq.html?"';
   put "|| :&AnnoAccession[selrow];";
   put 'web(url);wait(3);';
   put '));););';
   put " ";
	
   put 'OE = expr(';
   put '// --------------------------------------------------------------';
   put '// Open web browser,'; 
   put '// pre-set selected data table, column and path'; 
   put '// for creating and saving the query file.';
   put '// --------------------------------------------------------------';
   put 'url = "http://vortex.cs.wayne.edu:8080/ontoexpress/servlet/UserInfo";';
   put 'web(url); wait(2);';
   put 'dt = current data table();';
   put "col = char(expr( &AnnoAccession ));";
   put 'path = "C:\Program Files\SAS Institute\SASSDS\1.3\Client\Temp\queryFile.txt";';
   put " ";
   put '// ---------------------------------------------------------------';
   put '// Display a dialog to let user reset column and path';
   put '// for creating and saving the query file.';
   put '// ---------------------------------------------------------------';
   put 'result=dialog( title(" Message"),';
   put 'vlist(" ", "  A query file is required for searching with Onto-Express. Enter a location to save the query file. The directory must already exist.  ",';
   put '"        ",'; 
   put '" ", " ", line up(2,';
   put '" The query file path:", path=Edit Text(path),';
   put '" ", " ", " The accession number column:", col=Edit Text(col)),'; 
   put '" ", " ", HList("                                                                      ", button("OK"), button("Cancel"))));';
   put 'path = result["path"];';
   put 'col = result["col"];';
   put " ";
   put '// ----------------------------------------------------------------------';
   put '// Create and save subset data table from selected rows and column.';  
   put '// The subset data table will not be created if no row has been selected.';
   put '// ----------------------------------------------------------------------';
   put 'if ( nrow(currentDataTable()<<getSelectedRows)>0,';
   put 'Query = dt << Subset(Output Table("queryFile"), Linked, Selected rows, columns(eval(col))),';
   put 'dialog( title(" Warning"), vlist( " ", " No row has been selected.",';
   put '" ", " The query file was not generated.", " ", " ",'; 
   put 'HList("              ", button("OK")))););';
   put 'Query << Save (eval(path), Preferences(Export Settings(Export Table Headers(0))));';
   put 'dt<<clear select;';
   put ');';
   put " ";

   put 'searchWindow = New Window("Web Search",vlistBox (';
   put 'ButtonBox("UniGene",ug),';
   put 'ButtonBox("GenBank-Nucleotide",GBN),';
   put 'ButtonBox("FlyBase",flyBaseSearch),';
   put 'ButtonBox("Onto-Express",OE),';
   put 'ButtonBox("Close",searchWindow<<CloseWindow) )';
   put ');';
   put " ";

	%end;

%mend WebSearchJSL;

