/*
SAS Microarray Solution Annotation Input Engine

Name: AnnotationAP
Author: Shibing Deng, SAS Institute Inc.
Support: Tzu-Ming Chu, SAS Institute Inc.
History:
   shdeng 26Feb04 initial coding

Description: Annotation AP imports user provided annotation file into SAS data set.
Then output an unigue ID to Java program for querying annotation.  Annotation AP then read in
the table created by java and merge back with user supplied annotation file.
*/
OPTIONS nospool;


*ProcessVariables;


%let AnnoInFile= ; * Input file with full path. The file should have an Gene ID column with column name given in GeneIDVar below;

%let AnnoType= ; * Input file type. It must be one of the following:
	  1. TAB = Tab delimited text file,
      2. CSV = comma delimited text file,
      3. EXCEL = Excel file (Excel 2000 or later) (not accepted for Unix server),
      4. SAS = SAS dataset.;
%let GeneIDVar= ; *Column or Variable name of Gene Identifier.;
%let GeneIdentifier= ; * Gene Identifier. It must be one of the following: 
      1. GenBank Accession,
      2. UniGene Cluster Id,
      3. Locus Link Id,  
      4. Gene Symbol. 3 and 4 will be implemented in next release ;


%let AnnoOutData = ; * Name of annotation in SAS data format;
%let OutPath = ; * Path the converted SAS data stored;
%let MacroPath = ; * Path the UtilityMacros stored;


*ProcessBody;

/**
%let AnnoInFile= ; * Input file with full path. The file should have an Gene ID column with column name given in GeneIDVar below;

%let AnnoType= ; * Input file type. It must be one of the following:
	  1. TAB = Tab delimited text file,
      2. CSV = comma delimited text file,
      3. EXCEL = Excel file (Excel 2000 or later),
      4. SAS = SAS dataset.;
%let GeneIDVar= ; *Column or Variable name of Gene Identifier.;
%let GeneIdentifier= ; * Gene Identifier. It must be one of the following: 
      1. GenBank Accession,
      2. UniGene Cluster Id,
      3. Locus Link Id,  
      4. Gene Symbol. 3 and 4 will be implemented in next release ;
%let Organism= ; * Organism. It must be one of the following:
	  1. Homo sapiens,
	  2. Mus musculus,
	  3. Rattus norvegicus,
	  4. Drosophila melanogaster.;

%let accSelected= TRUE; * true/false indicating whether user wants GenBank Accession as part of anno info;
%let clustTitleSelected= TRUE; * true/false indicating whether user wants UniGene Title as part of anno info;
%let clustIdSelected= TRUE; * true/false indicating whether user wants UniGene Cluster Id & link as part of anno info;
%let symbolSelected= TRUE; * true/false indicating whether user wants Gene Symbol & link as part of anno info;
%let locusIdSelected= TRUE; * true/false indicating whether user wants Locus Id & link as part of anno info;
%let entrezGeneSelected= TRUE; * true/false indicating whether user wants Entrez Gene link as part of anno info;
%let homolSelected= TRUE; * true/false indicating whether user wants HomoloGene link as part of anno info;
%let chromoSelected= TRUE; * true/false indicating whether user wants Chromosome Location as part of anno info;
%let cytoSelected= TRUE; * true/false indicating whether user wants Cytogenetic Location as part of anno info;
%let giSelected= TRUE; * true/false indicating whether user wants GI Number & link as part of anno info;
%let protIdSelected= TRUE; * true/false indicating whether user wants Protein Id & link as part of anno info;
%let cloneIdSelected= TRUE; * true/false indicating whether user wants Clone Id as part of anno info;
%let libIdSelected= TRUE; * true/false indicating whether user wants Library Id & link as part of anno info;
%let seqTypeSelected= TRUE; * true/false indicating whether user wants Sequence Type as part of anno info;
%let keggGeneSelected= TRUE; * true/false indicating whether user wants KEGG Gene link as part of anno info;
%let omimSelected = TRUE; * true/false indicating whether user wants OMIM Id & link as part of anno info;
%let swissSelected = TRUE; * true/false indicating whether user wants Swissprot & link as part of anno info;
%let ensembSelected= TRUE; * true/false indicating whether user wants Ensembl link as part of anno info;
%let nciSelected= TRUE; * true/false indicating whether user wants CGAP Pathway link as part of anno info;
%let goSelected= TRUE; * true/false indicating whether user wants Gene Ontology Term and link as part of anno info;
%let pubmedCitSelected = TRUE; * true/false indicating whether user wants Pubmed/Citations as part of anno info;
%let pubmedGeneSelected= TRUE; * true/false indicating whether user wants Pubmed/Gene Name as part of anno info;
%let pubmedTransgenSelected= TRUE; * true/false indicating whether user wants Pubmed/Transgene/Knockout as part of anno info;
%let entrezSelected= TRUE; * true/false indicating whether user wants Entrez link as part of anno info;
%let organismSelected= TRUE; * true/false indicating whether user wants Organism name as part of anno info;
%let htmlOneRowPerRecord=Multiple rows per gene; * Value indicating whether user selects one row per gene display or multiple row per gene display in Anno HTML file;
%let cloneEndSelected=TRUE; * true/false indicating whether user wants 3' or 5' End of clone info as part of anno info;
%let refSeqSelected=TRUE; * true/false indicating whether user wants reference sequence Accessions as part of anno info;
%let cgapGeneSelected=TRUE; * true/false indicating whether user wants CGAP Gene link as part of anno info;

%let AnnoOutData = ; * Name of annotation in SAS data format;
%let OutPath = ; * Path the converted SAS data stored;
%let MacroPath = ; * Path the UtilityMacros stored;
%let SDS_Create_Anno = true; * ;
**/

*Internal macro variable;
%let AnnoStep=1;*Step in creating Annotation file, value= 1 or 2;

%macro AnnoParse;

%let outTxtFileName = &AnnoOutData..txt;
%let outHtmlFileName = &AnnoOutData..html;

*initialization;
%let ProcessName = AnnotationAP;

%global exiterror;
%let exiterror = 0;
%let clockrunning = 0;
options nospool;
data _null_;
   infile "&MacroPath.UtilityMacros.sas";
run;
%if &syserr %then %do;
   %put ERROR: Path for SAS macros is invalid.;
   %goto exit;
%end;
%include "&MacroPath.UtilityMacros.sas" / nosource;
%ClockStart(&ProcessName);
%if &syserr %then %goto exit;
%let clockrunning = 1;

%CheckDataName(_name=AnnoOutData,_default=annooutdata);

%CheckForBlank(&AnnoInFile,Annotation input File);
%if &exiterror %then %goto exit;
%PathName(&AnnoInFile);
%if &exiterror %then %do;
   %put ERROR: Annotation input dataset specification is invalid.;
   %goto exit;
%end;
%CheckForBlank(&AnnoType,Data type of the annotation input file);
%if &exiterror %then %goto exit;
%CheckExcel(&AnnoType,the annotation input file);
%if &exiterror %then %goto exit;
%let AnnoType=%upcase(&AnnoType);
%if &annotype=1 | &annotype=TABLE | &annotype=TXT %then %let annotype=TAB;
%if &annotype=2  %then %let annotype=CSV;
%if &annotype=3 | &annotype=EXCEL | &annotype=XLS %then %let annotype=EXCEL2000;
%if ^(&AnnoType = TAB | &AnnoType=CSV | &AnnoType=SAS | &AnnoType=EXCEL2000) %then %do;
	%put ERROR: Incorrect AnnoType.;
   %goto exit;
%end;
%CheckForBlank(&OutPath,Output path);
%if &exiterror %then %goto exit;
%CheckForBlank(&GeneIDVar,Output path);
%if &exiterror %then %goto exit;
%CheckDataName(_name=GeneIDVar,_default=accession);

%CheckForBlank(&AnnoOutData,Annotation Output Dataset);
%if &exiterror %then %goto exit;
%*if ^%length(&GeneIDVar) %then %let GeneIDVar=ACCESSION;/*default to accession*/

* define libnames;
Libname AnnoLib "&OutPath";

%if &sysscp=WIN %then %let _slash=\;
%else %let _slash=/;

%if &AnnoType = SAS %then %do;
   %PathName(&AnnoInFile);
   libname TmpLib "&TmpPath";

   data annolib.&AnnoOutData;
      set Tmplib.&TmpName;
	  /*%if %length(&GeneIDVar) %then %do;rename &GeneIDVar=GeneID;%end;*/
   run;
%end;
%else %do;
   %let workpath=%sysfunc(pathname(work));
   proc registry export="&workpath.&_slash.registry.txt";
   run;
   %ImportScanRows(Flag=1,ScanRows=200);
   PROC IMPORT OUT= annolib.&AnnoOutData
        DATAFILE= "&AnnoInFile"
        DBMS=&AnnoType REPLACE;
        GETNAMES=YES;
   RUN;
   %if %sysfunc(fexist(registry)) %then %do;
      filename registry "&workpath.&_slash.registry.txt";
      proc registry import=registry;
      run;
      %let frc=%sysfunc(fdelete(registry));
   %end;
%end;

%NObsVars(annolib.&AnnoOutData);
%if (&nobs=0 | &nvars=0) %then %do;
   data annolib.exitedwitherrors;
      error=1;
   run;
   %put ERROR: Error(s) were encountered creating annotation output data set.;
   %let exiterror = 1;
   %goto exit;
%end;

%CheckVarExist(&GeneIDVar,annolib.&AnnoOutData,Gene Identifier Variable);

   data annolib.&AnnoOutData;
      set annolib.&AnnoOutData;
	  %if %upcase(&GeneIdentifier) = GENBANK ACCESSION %then %do;
         &GeneIDVar = scan(&GeneIDVar,1,'.');
	  %end;
	  %else %if %upcase(&GeneIdentifier) = UNIGENE CLUSTER ID %then %do;
	     if substr(&GeneIDVar,1,1) = '-' then &GeneIDVar = ' ';
	     else &GeneIDVar = scan(&GeneIDVar,1,'/'); 
	  %end;
   run;

  *Elliminate Dup ID;
   proc sort data=annolib.&AnnoOutData;
      by &GeneIDVar;
   run;

   data _null_;
	  set annolib.&AnnoOutData;
	  by &GeneIDVar;
	  file "&outpath.GeneID4java.txt";
	  if _n_=1 then put "&GeneIDVar";
	  if first.&GeneIDVar;
	  put &GeneIDVar;
   run;



%exit:;
%if (&syserr | &exiterror) %then %put ERROR: &ProcessName exited due to errors.;
%if (&clockrunning) %then %ClockStop(&ProcessName);

%mend AnnoParse;
%AnnoParse;

